/* 
 * Copyright (C) 2007 Laird Breyer
 *  
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA.
 * 
 * Author:   Laird Breyer <laird@lbreyer.com>
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <math.h>
#include <ctype.h>
#include <string.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <time.h>

#if defined HAVE_UNISTD_H
#include <unistd.h> 
#include <sys/types.h>
#include <sys/wait.h>
#endif

#include "util.h"
#include "spherecl.h" /* make sure this is last */

/* global variables */

extern options_t u_options;
extern options_t m_options;
extern charparser_t m_cp;
extern digtype_t m_dt;
extern char *extn;

extern char *textbuf;
extern charbuf_len_t textbuf_len;

/* for option processing */
extern char *optarg;
extern int optind, opterr, optopt;

extern char *progname;
extern char *inputfile;
extern long inputline;

extern category_count_t cat_count;

extern int cmd;
int exit_code = 0; /* default */

extern long system_pagesize;

extern void *in_iobuf;
extern void *out_iobuf;


char *classifier_argv[] = {
  "-m",
  "-v",
  "-Y"
};
int classifier_argc = sizeof(classifier_argv)/sizeof(char *);

char *learner_argv[] = {
  "-H", "25",
  "-T", "email",
  "-T", "email:headers",
  "-T", "email:theaders",
  "-T", "email:xheaders",
  "-L", "uniform",
  "-e", "cef2"
};
int learner_argc = sizeof(classifier_argv)/sizeof(char *);


/***********************************************************
 * MISCELLANEOUS FUNCTIONS                                 *
 ***********************************************************/

static void usage(/*@unused@*/ char **argv) {
  fprintf(stderr, 
	  "\n");
  fprintf(stderr, 
	  "spherecl [OPTIONS] -C CATEGORY1:CATEGORY2 [FILE]...\n");
  fprintf(stderr, 
	  "\n");
  fprintf(stderr, 
	  "      classifies FILE as either CATEGORY1 or CATEGORY2.\n");
  fprintf(stderr, 
	  "\n");
  fprintf(stderr, 
	  "spherecl [OPTIONS] -C CATEGORY1:CATEGORY2 [-l CATEGORY] [FILE]...\n");
  fprintf(stderr, 
	  "\n");
  fprintf(stderr, 
	  "      learns each FILE as an example of CATEGORY.\n");
  fprintf(stderr, 
	  "\n");
  fprintf(stderr, 
	  "spherecl -V\n");
  fprintf(stderr, 
	  "\n");
  fprintf(stderr, 
	  "      prints program version.\n");
}

/***********************************************************
 * MAIN FUNCTIONS                                          *
 ***********************************************************/

/* if path contains progname, replace with what. Else just use what and
   hope the system path will find it */
char *get_dbacl_path(const char *arg0) {
  const char *what = "dbacl";
  char *q = NULL;
  char *d = (char *)what;

  if( arg0 && (q = strstr(arg0, progname)) &&
      (d = strdup(arg0)) ) {
    strncpy(d + (q - arg0), what, strlen(progname));
    d[q - arg0 + strlen(what)] = '\0';
  }
  return d;
}


int set_option(int op, char *optarg) {
  int c = 0;
  switch(op) {
  case 'V':
    fprintf(stdout, "spherecl version %s\n", VERSION);
    fprintf(stdout, COPYBLURB, "spherecl");
    exit(0);
    break;
  case 'C':
    if( u_options & (1<<U_OPTION_CLASSIFY) ) {
      errormsg(E_ERROR, "-C switch can only be used once\n");
      exit(1);
    } else if( !*optarg || !strchr(optarg, ':') ) {
      errormsg(E_ERROR, "-C switch requires two category names separated by a colon\n");
    } 
    u_options |= (1<<U_OPTION_CLASSIFY);    
    c++;
    break;
  case 'l': 
    if( !(u_options & (1<<U_OPTION_CLASSIFY)) ) {
      errormsg(E_ERROR,
	       "please use the -C switch before the -l switch\n");
      exit(1);
    } else if( u_options & (1<<U_OPTION_LEARN) ) {
      errormsg(E_ERROR,
	       "option -l can only occur once\n");
      exit(1);
    } else if( !*optarg ) {
      errormsg(E_ERROR, "category name must not be empty\n");
    } else {

    }
    c++;
    break;
  case '@':
    /* this is a boundary marker. It indicates that some options
       are passed directly to the dbacl process. Note that dbacl
       will ignore '-@' as a NOOP */
    u_options |= (1<<U_OPTION_APPEND);
    break;
  default:
    c--;
    break;
  }
  return c;
}

void sanitize_options() {

  /* consistency checks */
  if( ((u_options>>U_OPTION_CLASSIFY) & 1) + 
      ((u_options>>U_OPTION_LEARN) & 1) != 1 ) {
    errormsg(E_ERROR, "please use either -c or -l option.\n");
    exit(1);
  }

}

char **extend_argv(int argc, char **argv, int xargc, char **xargv) {
  int i;
  char **xv;
  /* calloc -> no need to NULL terminate array */
  xv = calloc(argc+xargc, sizeof(char *));
  if( xv ) {
    /* xargs always come first */
    for(i = 0; i < xargc; i++) {
      xv[i] = xargv[i];
    }
    /* skip argv[0] which contains pathname, also implies
     last element of xv is always NULL */
    for(i = 0; i < argc; i++) {
      xv[xargc + i] = argv[i+1];
    }
  }
  return xv;
}


int run_process(const char *path, char **argv) {
  pid_t pid;
  /* two pipes parent to child and child to parent */
  int p2c[2], c2p[2];
  int n, extra_lines = 0;

  if( (pipe(p2c) < 0) || (pipe(c2p) < 0) ) {
    errormsg(E_FATAL, "can't open pipes\n");
  }

  if( (pid = fork()) < 0 ) {
    errormsg(E_FATAL, "couldn't fork %s\n", path);
  } else if( pid == 0 ) {
    /* child reads from p2c and writes to c2p */
    close(p2c[1]); /* no writing to p2c */
    close(c2p[0]); /* no reading from c2p */

    /* make sure p2c is standard input */
    if( p2c[0] != STDIN_FILENO ) {
      if( dup2(p2c[0], STDIN_FILENO) != STDIN_FILENO ) {
	errormsg(E_FATAL, "p2c dup2 error\n");
      }
      close(p2c[0]);
    }

    /* make sure c2p is standard output */
    if( c2p[1] != STDOUT_FILENO ) {
      if( dup2(c2p[1], STDOUT_FILENO) != STDOUT_FILENO ) {
	errormsg(E_FATAL, "c2p dup2 error\n");
      }
      close(c2p[1]);
    }

    /* now replace child with new process */
    if( 0 > execvp(path, argv) ) {
      errormsg(E_FATAL, "couldn't execute %s\n", path);
    }
    
    /* we can't get here */
    exit(0);

  } else if( pid > 0 ) {
    /* parent writes to p2c and reads from c2p */
    close(p2c[0]); /* no reading from p2c */
    close(c2p[1]); /* no writing to c2p */
    
    if( !(u_options & (1<<U_OPTION_STDIN)) ) {
      /* now parent writes all of stdin to p2c, then closes p2c */
      /* if we fail to write to the child, it's probably dead, but we
	 keep going regardless */
      /* note that this is probably not the fastest stream copy */
      while( fill_textbuf(stdin, &extra_lines) ) {
	n = strlen(textbuf);
	if( write(p2c[1], textbuf, n) != n ) {
	  errormsg(E_WARNING, 
		   "unable to pipe some data, results may be incorrect\n");
	  break; /* assume we're done */
	}
      }
    }
    close(p2c[1]);

  }
  /* c2p[0] is still open and will have to be closed after it is read */
  return c2p[0];
}

size_t capture_textbuf(int fd) {
  size_t n;
  char *s;
  
  /* incorrect! if the child isn't ready, we think the data is done */
  s = textbuf;
  do {
    n = read(fd, s, textbuf_len - (s - textbuf));
    if( n < 0 ) {
      exit(1);
      errormsg(E_WARNING, "incomplete return data\n");
      break;
    } else if( n > 0 ) {
      s += n;
      if( s >= textbuf + textbuf_len ) {
	s = (char *)realloc(textbuf, 2 * textbuf_len);
	if( !s ) {
	  errormsg(E_FATAL, "not enough memory (runaway process?)\n"); 
	}
	textbuf = s;
	s += textbuf_len;
	textbuf_len *= 2;
      }
    }
  } while( n > 0 );

  /* make a string. if we're here, then s < textbuf + textbuf_len or
     we had a fatal error */
  *s = '\0';
  return (s - textbuf);
}

/* returns -1 on problem or process exit status otherwise */
int spawn_dbacl(int argc, char **argv,
		int xargc, char **xargv) {
  char *dbacl_path = "";
  int status;
  int fdo;
  char **xv;

  dbacl_path = get_dbacl_path(argv[0]);
  if( *dbacl_path ) {
    xv = extend_argv(argc, argv, xargc, xargv);
    if( xv ) {
      fdo = run_process(dbacl_path, xv);
      free(xv);
      if( fdo > -1 ) {
	/* captures complete output in textbuf, not null terminated */
	capture_textbuf(fdo);
	close(fdo);
	wait(&status); /* find out if child ran ok */
	return WIFEXITED(status) ? WEXITSTATUS(status) : -1;
      }
    }
  }
  return -1;
}

bool_t spawn_classifier(int argc, char **argv) {
  spawn_dbacl(argc, argv, classifier_argc, classifier_argv);
  printf("%s", textbuf);
  return (bool_t)1;
}

bool_t spawn_learner(int argc, char **argv) {
  spawn_dbacl(argc, argv, learner_argc, learner_argv);
  printf("%s", textbuf);
  return (bool_t)1;
}

int main(int argc, char **argv) {

  signed char op;

  progname = "spherecl";
  inputfile = "stdin";
  inputline = 0;

#if defined(HAVE_GETPAGESIZE)
  system_pagesize = getpagesize();
#endif
  if( system_pagesize == -1 ) { system_pagesize = BUFSIZ; }

  init_signal_handling();

  /* parse the options */
  while( (op = getopt(argc, argv, 
		      "l:C:V@")) > -1 ) {
    set_option(op, optarg);
    if( u_options & (1<<U_OPTION_APPEND) ) {
      /* -@ detected, stop processing remaining arguments */
      break;
    }
  }

  /* end option processing */
  sanitize_options();

  if( (optind > -1) && *(argv + optind) && !(cmd & (1<<CMD_QUITNOW)) ) {
    u_options |= (1<<U_OPTION_STDIN);
  }

  init_buffers();
  
  /* set up callbacks */
  if( u_options & (1<<U_OPTION_CLASSIFY) ) {
    spawn_classifier(argc, argv);
  } else if( u_options & (1<<U_OPTION_LEARN) ) {
    /* category learning */
    spawn_learner(argc, argv);
  } else { /* something wrong ? */
    usage(argv);
    exit(1);
  }

  cleanup_buffers();

  cleanup_signal_handling();

  return exit_code;
}


